"""Module dedicated to type conversions."""

import ast
import re
from datetime import datetime
from decimal import Decimal, localcontext
from typing import Any, Callable, Optional, Type, Union


class InvalidValueException(Exception):
    """Exception thrown when value cannot be parsed to specified data type."""

    pass


class BaseType:
    """Base class for data type."""

    subclasses: list = []
    convert: Union[Callable[..., Any], Type[Any]]
    dtype: str

    def __init_subclass__(cls: Any, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        cls.subclasses.append(cls)


def get_type(dtype: str) -> BaseType:
    """Gets Data Type object for provided string representation.

    Args:
        dtype (str): String representation of data type.

    Returns:
        Optional[BaseType]: Object representing data type.
    """
    for factory in BaseType.subclasses:
        current = factory.get_for_type(dtype)
        if current:
            return current
    raise InvalidValueException(f"Not supported: {dtype}")


class IntType(BaseType):
    convert = int

    @staticmethod
    def get_for_type(dtype: str) -> Optional[BaseType]:
        return IntType(dtype.lower()) if dtype.lower() in ("int", "bigint") else None

    def __init__(self, dtype: str):
        self.dtype = dtype


class FloatType(BaseType):
    convert = float

    def __init__(self, dtype: str):
        self.dtype = dtype

    @staticmethod
    def get_for_type(dtype: str) -> Optional[BaseType]:
        return FloatType(dtype.lower()) if dtype.lower() in ("float", "double") else None


class StringType(BaseType):
    convert = str
    dtype = "string"

    @staticmethod
    def get_for_type(dtype: str) -> Optional[BaseType]:
        return StringType() if dtype.lower() in ("string", "str") else None


class BooleanType(BaseType):
    dtype = "boolean"

    @staticmethod
    def get_for_type(dtype: str) -> Optional[BaseType]:
        return BooleanType() if dtype.lower() in ("boolean", "bool") else None

    def convert(self, value: str) -> bool:
        return value.lower() == "true"


class DateTimeType(BaseType):
    dtype = "timestamp"

    @staticmethod
    def get_for_type(dtype: str) -> Optional[BaseType]:
        return DateTimeType() if dtype.lower() == "timestamp" else None

    def convert(self, value: str) -> datetime:
        return datetime.strptime(value, "%Y-%m-%d %H:%M:%S")


class DecimalType(BaseType):
    @staticmethod
    def get_for_type(dtype: str) -> Optional[BaseType]:
        lower_type = dtype.lower()
        if lower_type.startswith("decimal"):
            precision, scale = map(int, re.split(r"\.|,", lower_type.replace("decimal", "")[1:-1]))
            return DecimalType(precision, scale)
        return None

    def __init__(self, precision: int, scale: int):
        self.precision = precision
        self.scale = scale
        self.dtype = f"decimal({precision},{scale})"

    def convert(self, value: str) -> Decimal:
        if len(str(int(float(value)))) <= (self.precision - self.scale):
            with localcontext() as ctx:
                ctx.prec = self.precision
                return Decimal(value)
        else:
            raise InvalidValueException(
                "{} not in range {}".format(value, (self.precision - self.scale) * "9")
            )


class MapType(BaseType):
    @staticmethod
    def get_for_type(dtype: str) -> Optional[BaseType]:
        lower_type = dtype.lower()
        if lower_type.startswith("map"):
            matches = re.match(r"map<(\w+),\s?(.+)>", dtype)
            if not matches:
                return None

            key_type = matches.group(1)
            value_type = matches.group(2)
            return MapType(key_type, value_type)
        return None

    def __init__(self, key_type: str, value_type: str):
        self.key_type = get_type(key_type)
        self.value_type = get_type(value_type)
        self.dtype = f"map<{self.key_type.dtype},{self.value_type.dtype}>"

    def convert(self, value: str) -> dict:
        items = ast.literal_eval(value)
        result = {}
        for key, val in items.items():
            result[self.key_type.convert(key)] = self.value_type.convert(str(val))
        return result


class ArrayType(BaseType):
    @staticmethod
    def get_for_type(dtype: str) -> Optional[BaseType]:
        lower_type = dtype.lower()
        if lower_type.startswith("array"):
            matches = re.match(r"array<(.+)>", dtype)
            if not matches:
                return None
            val_type = matches.group(1)
            return ArrayType(val_type)
        return None

    def __init__(self, value_type: str):
        self.value_type = get_type(value_type)
        self.dtype = f"array<{self.value_type.dtype}>"

    def convert(self, value: str) -> list:
        items = ast.literal_eval(value)
        result = []
        for val in items:
            result.append(self.value_type.convert(val))
        return result



"""Module dendicated for list conversions to data structures, needed for DataFrame construction"""
from typing import Any, Dict, List, Tuple

from markdown_frames.type_definitions import get_type


class InvalidDataException(Exception):
    """Exception thrown when Invalid table input is provided."""

    pass


class Column:
    """DataFrame column representation."""

    def __init__(self, name: str, dtype: str):
        self.name = name
        self._converter = get_type(dtype)
        self.dtype = self._converter.dtype if self._converter else dtype

    def value(self, val: str) -> Any:
        """Converts provided value to column data type.

        Args:
            val (str): String representation of the value.

        Returns:
            Any: Value converted to columns' data type.
        """
        if val is None or val.lower() in ("none", "null", "na"):
            return None

        return self._converter.convert(val)


class Table:
    """Table representation for DataFrame"""

    def __init__(self, table_rows: List[List[str]]):
        if len(table_rows) < 1:
            raise InvalidDataException("Table must contain at least 2 rows - header and types")
        self._column_names = table_rows[0]
        self._column_types = table_rows[1]
        self._data = table_rows[2:]

    def _columns(self) -> List[Column]:
        return [Column(name, dtype) for name, dtype in zip(self._column_names, self._column_types)]

    def to_pandas_data(self) -> Dict[str, List[Any]]:
        """Converts Table to data needed for Pandas DataFrame.

        Returns:
            dict: Dict, where each key is column value, and value - list of values.
        """
        result: Dict[str, List[Any]] = {}
        columns = self._columns()
        for row in self._data:
            for idx, val in enumerate(row):
                col = columns[idx]
                value = col.value(val)
                if col.name in result:
                    result[col.name].append(value)
                else:
                    result[col.name] = [value]
        return result

    def to_spark_data(self) -> Tuple[List[Tuple[Any, ...]], str]:
        """Converts Table to data needed for Spark DataFrame.

        Returns:
            Tuple: Tuple where first element is list of tuples representing row,
                and second element is Spark schema representation string.
        """
        columns = self._columns()
        data = []
        columns = self._columns()
        for row in self._data:
            data.append(tuple([columns[idx].value(val) for idx, val in enumerate(row)]))

        schema = ", ".join([f"{col.name} {col.dtype}" for col in columns])
        return (data, schema)



"""Functions that parse markdown table or CSV file to Apache Spark (PySpark)
DataFrame.
"""
from pyspark.sql import DataFrame, SparkSession

from markdown_frames.table import Table
from markdown_frames.utils import parse_file, parse_markdown


def spark_df(markdown: str, spark: SparkSession) -> DataFrame:
    """Returns a Spark DataFrame with specified types.

    Args:
        markdown (str): Markdown representation of input data.
        spark (SparkSession): Spark Session object.

    Returns:
        DataFrame: DataFrame with data and schema specified.
    """

    table = Table(parse_markdown(markdown))
    data, schema = table.to_spark_data()
    return spark.createDataFrame(data, schema)


def spark_df_from_csv(path: str, spark: SparkSession, delimiter: str = ",") -> DataFrame:
    """Returns a Spark DataFrame built from CSV. Second line in CSV must represent data types.

    CSV file format
    ```
    column_name1, column_name2, column_name3
    string, int, float
    hellow, 1, 1.1
    goodbye, 2, 2.2
    ```
    Args:
        path (str): Path where `.csv` file is placed.

    Returns:
        DataFrame: Spark DataFrame for provided path.
    """

    table = Table(parse_file(path, delimiter))
    data, schema = table.to_spark_data()
    return spark.createDataFrame(data, schema)



"""Functions that parse markdown table or CSV file to Pandas DataFrame."""

import pandas as pd

from markdown_frames.table import Table
from markdown_frames.utils import parse_file, parse_markdown


def pandas_df(markdown_table: str) -> pd.DataFrame:
    """Returns a Pandas DataFrame with specified types.

    Args:
        markdown_table (str): Markdown representation of input data.

    Returns:
        DataFrame: Pandas Dataframe with data.
    """

    table = Table(parse_markdown(markdown_table))
    return pd.DataFrame(data=table.to_pandas_data())


def pandas_df_from_csv(path: str, delimiter: str = ",") -> pd.DataFrame:
    """Returns a Pandas DataFrame built from CSV. Second line in CSV must represent data types.

    CSV file format
    ```
    column_name1, column_name2, column_name3
    string, int, float
    hellow, 1, 1.1
    goodbye, 2, 2.2
    ```
    Args:
        path (str): Path where `.csv` file is placed.

    Returns:
        DataFrame: Pandas DataFrame for provided path.
    """

    table = Table(parse_file(path, delimiter))
    return pd.DataFrame(data=table.to_pandas_data())



"""Functions that are used for parsing."""

import csv
from typing import List

_SEPARATOR_SYMBOLS = ("#", "-", "+", "=", "*")


def _parse_row(row_string: str) -> List[str]:
    """Returns a formatted values for a row.

    Args:
        row_string (str): String that represent one row for input table.

    Returns:
        List[str]: :ist of formated values in the table row.
    """
    row_values = filter(lambda s: s.strip() != "", row_string.split("|"))
    return list(map(lambda s: s.strip(), row_values))


def parse_markdown(markdown_table: str) -> List[List[str]]:
    """Given markdown table produces a list of lists - table.

    Args:
        markdown_table (str): Table in markdown format.

    Returns:
        List[List[str]]: List of lists with rows of data (still in str format).
    """
    rows = map(_parse_row, markdown_table.split("\n"))
    # Remove empty rows
    filtered_table = list(filter(lambda x: x, rows))

    # Remove separator, if present
    if len(filtered_table) > 2:
        first_symbol = filtered_table[2][0][0]
        if first_symbol in _SEPARATOR_SYMBOLS and all(
            not len(val.replace(first_symbol, "")) for val in filtered_table[2]
        ):
            del filtered_table[2]

    return filtered_table


def parse_file(path: str, delimiter: str = ",") -> List[List[str]]:
    """Reads CSV file and returns Table representation for it.

    Args:
        path (str): File path.
        delimiter (str): CSV delimiter (default: `,`)

    Returns:
        List[List[str]]: List of lists with rows of data (still in str format).
    """
    with open(path, encoding="utf-8") as file:
        csv_file = csv.reader(file, delimiter=delimiter)
        return [list(map(lambda s: s.strip(), row)) for row in csv_file]



"""Setup script for the markdown parsing library."""

from distutils.core import setup
from pathlib import Path

from setuptools import find_packages

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="markdown_frames",
    version="1.0.3",
    packages=find_packages(
        exclude=(
            "tests",
            "docs",
        )
    ),
    install_requires=[],
    author="Exacaster",
    author_email="support@exacaster.com",
    url="https://github.com/exacaster/markdown_frames",
    description="Markdown tables parsing to pyspark / pandas DataFrames",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Testing",
    ],
    extras_require={
        "pyspark": ["pyspark"],
        "pandas": ["pandas"],
    },
)
